"""Status icon"""

from PyQt5 import QtCore, QtGui, QtWidgets

from .preferencesdialog import PreferencesDialog
from .aboutdialog import AboutDialog
from .paths import PICS_PATHS


class StatusIcon(QtWidgets.QSystemTrayIcon):
    """Status icon"""

    def __init__(self, main_window, settings):

        # TODO: enlarge icon to best fit
        super().__init__(QtGui.QIcon(PICS_PATHS["birthday"]), main_window)

        self.main_window = main_window
        self.bday_list = main_window.bday_list
        self.settings = settings

        # TODO: Is there a standard translation in Qt for those menu actions?
        menu = QtWidgets.QMenu()
        menu.addAction(
            QtGui.QIcon.fromTheme("view-refresh"),
            self.tr("Refresh"),
            self.main_window.reload,
        )
        menu.addAction(
            QtGui.QIcon.fromTheme("preferences-other"),
            self.tr("Preferences"),
            lambda: PreferencesDialog(self.settings, self.main_window).exec_(),
        )
        menu.addAction(
            QtGui.QIcon.fromTheme("help-about"),
            self.tr("About"),
            lambda: AboutDialog(self.main_window).exec_(),
        )
        menu.addAction(
            QtGui.QIcon.fromTheme("application-exit"),
            self.tr("Quit"),
            QtCore.QCoreApplication.instance().quit,
        )

        # Set context menu to open on right click
        self.setContextMenu(menu)

        # Display birthdays on left click
        def tray_icon_activated_cb(reason):
            if (
                reason == QtWidgets.QSystemTrayIcon.Trigger
                or reason == QtWidgets.QSystemTrayIcon.DoubleClick
            ):
                # Toggle birthday window visibility
                self.main_window.setVisible(not self.main_window.isVisible())

        self.activated.connect(tray_icon_activated_cb)

        self._show_when_systray_available()

    def _show_when_systray_available(self):
        """Show status icon when system tray is available

        If available, show icon, otherwise, set a timer to check back later.
        This is a workaround for https://bugreports.qt.io/browse/QTBUG-61898
        """
        if self.isSystemTrayAvailable():
            self.show()
        else:
            QtCore.QTimer.singleShot(1000, self._show_when_systray_available)

    def reload_set_icon(self):
        """Set icon according to birthday status"""
        if self.bday_list.bdays_in_period():
            if self.bday_list.check_day(0):
                # Birthday today
                self.setIcon(QtGui.QIcon(PICS_PATHS["birthdayred"]))
            else:
                self.setIcon(QtGui.QIcon(PICS_PATHS["birthday"]))
        else:
            self.setIcon(QtGui.QIcon(PICS_PATHS["nobirthday"]))



"""BirthdayList module

BirthdayList is a structure storing birthdates and keeping track of birthdays
in a specified period around today's date.
"""

import datetime as dt
from collections import defaultdict

from PyQt5 import QtCore

from .ics_export import ICSExport
from .backends import BACKENDS
from .backends.exceptions import BackendReadError
from .exceptions import ICSExportError


class BirthdayList(QtCore.QObject):
    """Structure storing all birthdates and birthdays in specified period"""

    def __init__(self, main_window, settings):

        super().__init__()

        self.main_window = main_window
        self.settings = settings
        self.ics_export = ICSExport(settings)

        # dict storing all birthdates
        # key: dt.date
        # value: list of names
        self._birthdates = defaultdict(list)

        # dict storing birthdays in specified period
        # key: int (index of day in period)
        # value: list of (birthdate, name) tuples
        self._birthdays = defaultdict(list)

        self._needs_update = False

    def add(self, name, birthdate):
        """Add a new birthdate to the collection

        name (str): Person name
        birthdate (datetime.date): Date of birth
        """
        self._needs_update = True
        self._birthdates[birthdate].append(name)

    def bdays_in_period(self):
        """Return True, if there is a birthday in specified period"""
        first_day = self.settings.value("firstday", type=int)
        last_day = self.settings.value("lastday", type=int)
        for day in range(first_day, last_day + 1):
            if day in self._birthdays:
                return True
        return False

    def check_day(self, day_num=0):
        """Return all birthdays for a given day index in specified period"""
        if self._needs_update:
            self._update()
            self._needs_update = False
        return self._birthdays[day_num]

    def reload(self):
        """Reload birthdates from all backends and update birthdays"""

        self._birthdates.clear()

        for bcknd in BACKENDS:
            if bcknd.cls is not None and self.settings.value(
                bcknd.id + "/enabled", type=bool
            ):
                bcknd_inst = bcknd.cls(self.settings)
                try:
                    birthdates = bcknd_inst.parse()
                except BackendReadError as exc:
                    self.main_window.show_error_message(str(exc))
                else:
                    for name, date in birthdates:
                        self.add(name, date)

        self._update()

        if self.settings.value("ics_export/enabled", type=bool):
            try:
                self.ics_export.write(self._birthdates)
            except ICSExportError as exc:
                self.main_window.show_error_message(str(exc))

    def _update(self):
        """Update self.birthdays with all birthdays in specified period"""
        self._birthdays.clear()
        for day_num in range(
            self.settings.value("firstday", type=int),
            self.settings.value("lastday", type=int) + 1,
        ):
            day = dt.date.today() + dt.timedelta(day_num)
            for date, birthdate_list in self._birthdates.items():
                if day.day == date.day and day.month == date.month:
                    self._birthdays[day_num].extend(
                        [(date, name) for name in birthdate_list]
                    )



"""QBirthday

A KBirthday clone working with different backends
and relatively easy to extend to use other backends.

Original source from:
- pygtk-demo Status Icon: Nikos Kouremenos
- EvoBdayReminder.py: Axel Heim. http://www.axelheim.de/
"""

import sys

from PyQt5 import QtCore, QtGui, QtWidgets

from .mainwindow import MainWindow
from .paths import PICS_PATHS, QM_FILES_DIR


def main():
    """Load settings, start status icon and get to work."""

    # TODO: Think twice about naming before releasing
    QtCore.QCoreApplication.setOrganizationName("QBirthday")
    QtCore.QCoreApplication.setApplicationName("qbirthday")

    app = QtWidgets.QApplication([])
    app.setWindowIcon(QtGui.QIcon(PICS_PATHS["qbirthday"]))

    # Internationalization
    # Use system locale
    locale = QtCore.QLocale.system().name()
    # Load default translator for Qt strings
    translator_qt = QtCore.QTranslator()
    translator_qt.load(
        f"qt_{locale}",
        QtCore.QLibraryInfo.location(QtCore.QLibraryInfo.TranslationsPath),
    )
    app.installTranslator(translator_qt)
    # Load translator for own strings
    translator = QtCore.QTranslator()
    translator.load(str(QM_FILES_DIR / f"qbirthday_{locale}"))
    app.installTranslator(translator)

    # TODO: is this the right way?
    app.setQuitOnLastWindowClosed(False)

    # Main window
    MainWindow()

    sys.exit(app.exec_())



"""Settings module

Override QSettings to add default values
"""

from PyQt5 import QtCore

from .paths import GENERIC_DATA_LOCATION
from .backends import BACKENDS


CONFIG_DEFAULTS = {
    "firstday": -2,
    "lastday": 30,
}

ICS_EXPORT_CONFIG_DEFAULTS = {
    "enabled": False,
    "filepath": str(GENERIC_DATA_LOCATION / "qbirthday.ics"),
    "alarm": False,
    "alarm_custom_properties": "",
    "alarm_days": 1,
    "custom_properties": "",
}


class MissingDefaultError(Exception):
    """Missing parameter default value"""


class Settings(QtCore.QSettings):
    """QSettings child class with default values"""

    def setDefaultValue(self, key, value):  # pylint: disable=invalid-name
        """Set value unless it already exists"""
        self.setValue(key, self.value(key, value))

    def __init__(self):
        super().__init__()
        # General settings
        for key, val in CONFIG_DEFAULTS.items():
            self.setDefaultValue(key, val)
        # ICS export
        self.beginGroup("ics_export")
        for key, val in ICS_EXPORT_CONFIG_DEFAULTS.items():
            self.setDefaultValue(key, val)
        self.endGroup()
        # Backend settings
        for bcknd in BACKENDS:
            self.beginGroup(bcknd.id)
            # Disable all backends by default
            self.setDefaultValue("enabled", False)
            if bcknd.cls is not None:
                # If backend available, Load backend specific default values
                for key, val in bcknd.cls.DEFAULTS.items():
                    self.setDefaultValue(key, val)
            else:
                # If backend unavailable and enabled in config, warn user
                if self.value("enabled") is True:
                    # TODO: display warning
                    pass
            self.endGroup()



"""CSV file backend"""

import datetime as dt
import csv

from PyQt5 import QtCore, QtWidgets, uic

from qbirthday.paths import UI_FILES_DIR
from qbirthday.paths import APP_DATA_LOCATION
from .base import BaseBackend
from .exceptions import BackendReadError


BACKEND_ID = "CSV"
BACKEND_NAME = "CSV-file (comma separated value)"


class CSVPreferencesDialog(QtWidgets.QDialog):
    """CSV backend settings dialog"""

    def __init__(self, settings, parent):
        super().__init__(parent)
        uic.loadUi(str(UI_FILES_DIR / "csvpreferencesdialog.ui"), self)
        self.settings = settings
        self.filePathEdit.setText(self.settings.value("CSV/filepath"))
        self.filePathButton.clicked.connect(self.get_filepath)
        self.delimiterEdit.setText(self.settings.value("CSV/delimiter"))
        self.buttonBox.button(QtWidgets.QDialogButtonBox.Apply).clicked.connect(
            self.save
        )
        self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).clicked.connect(self.save)

        # TODO: disable OK and Apply if no path provided

    def get_filepath(self):
        """Get CSV file path"""
        filepath, _ = QtWidgets.QFileDialog.getOpenFileName(
            self, "Choose CSV file", self.filePathEdit.text() or QtCore.QDir.homePath()
        )
        if filepath:
            self.filePathEdit.setText(filepath)

    def save(self):
        """Save CSV backend settings"""
        self.settings.setValue("CSV/filepath", self.filePathEdit.text())
        self.settings.setValue("CSV/delimiter", self.delimiterEdit.text())


class CSVBackend(BaseBackend):
    """CSV file backend"""

    CONFIG_DLG = CSVPreferencesDialog

    DEFAULTS = {"filepath": str(APP_DATA_LOCATION / "qbirthday.csv"), "delimiter": ","}

    def __init__(self, settings):
        super().__init__(settings)
        self._filepath = self.settings.value("CSV/filepath", type=str)
        self._delimiter = self.settings.value("CSV/delimiter", type=str)

    def _raise_invalid_row_error(self, row):
        raise ValueError(
            self.tr('Invalid row "{row}" in CSV file {filepath}').format(
                row=self._delimiter.join(row), filepath=self._filepath
            )
        )

    def parse(self):
        """open and parse file"""

        birthdates = []

        try:
            with open(self._filepath) as csv_file:
                # Filter empty and commented rows
                filtered_rows = (
                    r for r in csv_file if (not r.startswith("#")) and r.strip()
                )
                for row in csv.reader(filtered_rows, delimiter=self._delimiter):
                    if len(row) != 2:
                        self._raise_invalid_row_error(row)
                    try:
                        date = dt.datetime.strptime(row[0], "%Y-%m-%d").date()
                    except ValueError:
                        self._raise_invalid_row_error(row)
                    else:
                        name = row[1].strip()
                        birthdates.append((name, date))
        except OSError:
            raise BackendReadError(
                self.tr("Can't open CSV file: {}").format(self._filepath)
            )
        except ValueError as exc:
            raise BackendReadError(exc)

        return birthdates


BACKEND = CSVBackend



"""iCalendar export"""

import datetime as dt

from PyQt5 import QtCore

from .exceptions import ICSExportError


class ICSExport(QtCore.QObject):
    """iCalendar export"""

    def __init__(self, settings):
        super().__init__()
        self.settings = settings

    def write(self, birthdates):
        """Write iCalendar file"""

        self.settings.beginGroup("ics_export")
        conf_filepath = self.settings.value("filepath", type=str)
        conf_alarm = self.settings.value("alarm", type=bool)
        conf_alarm_days = self.settings.value("alarm_days", type=int)
        conf_custom_properties = self.settings.value("custom_properties", type=str)
        conf_alarm_custom_properties = self.settings.value(
            "alarm_custom_properties", type=str
        )
        self.settings.endGroup()

        try:
            with open(conf_filepath, "w") as ics_file:
                now = dt.datetime.now().strftime("%Y%m%dT%H%M%SZ")
                ics_file.write("BEGIN:VCALENDAR\n")
                ics_file.write("VERSION:2.0\n")
                ics_file.write("PRODID:-//qbirthday//EN\n")
                for index, bday in enumerate(birthdates):
                    ics_file.write("BEGIN:VEVENT\n")
                    ics_file.write(f"UID:{now}-{index}@qbirthday\n")
                    ics_file.write(f"CREATED:{now}\n")
                    ics_file.write(f"LAST-MODIFIED:{now}\n")
                    ics_file.write(f"DTSTAMP:{now}\n")
                    ics_file.write("DTSTART:{}\n".format(bday.strftime("%Y%m%d")))
                    ics_file.write("DURATION:PT0S\n")
                    ics_file.write("CATEGORIES:{}\n".format(self.tr("Birthday")))
                    ics_file.write(
                        "SUMMARY:{}{}\n".format(
                            self.tr("Birthday: "), birthdates[bday][0]
                        )
                    )
                    ics_file.write("CLASS:PRIVATE\n")
                    ics_file.write("TRANSP:TRANSPARENT\n")
                    ics_file.write("RRULE:FREQ=YEARLY\n")
                    if conf_alarm:
                        ics_file.write("BEGIN:VALARM\n")
                        ics_file.write("ACTION:DISPLAY\n")
                        ics_file.write(f"TRIGGER;VALUE=DURATION:-P{conf_alarm_days}D\n")
                        ics_file.write(
                            "DESCRIPTION:{}{}\n".format(
                                self.tr("Birthday: "), birthdates[bday][0]
                            )
                        )
                        if conf_alarm_custom_properties != "":
                            ics_file.write(conf_alarm_custom_properties)
                            ics_file.write("\n")
                        ics_file.write("END:VALARM\n")
                    if conf_custom_properties != "":
                        ics_file.write(conf_custom_properties)
                        ics_file.write("\n")
                    ics_file.write("END:VEVENT\n")
                ics_file.write("END:VCALENDAR")
        except OSError:
            raise ICSExportError(
                self.tr(f"Can't write iCalendar file: {conf_filepath}")
            )



"""Preferences dialog"""

from pathlib import Path

from PyQt5 import QtWidgets, uic

from .backends import BACKENDS
from .paths import UI_FILES_DIR


class IcsExportPreferencesDialog(QtWidgets.QDialog):
    """ICS export settings dialog"""

    def __init__(self, settings, parent):

        super().__init__(parent)

        uic.loadUi(str(UI_FILES_DIR / "icsexportpreferencesdialog.ui"), self)

        self.settings = settings

        self.settings.beginGroup("ics_export")
        self.filePathEdit.setText(self.settings.value("filepath", type=str))
        self.alarmsCheckBox.setChecked(self.settings.value("alarm", False, type=bool))
        # TODO: add default values?
        self.alarmsDelaySpinBox.setValue(self.settings.value("alarm_days", type=int))
        self.customVeventTextEdit.setPlainText(
            self.settings.value("custom_properties", type=str)
        )
        self.customValarmTextEdit.setPlainText(
            self.settings.value("alarm_custom_properties", type=str)
        )
        self.settings.endGroup()

        self.filePathButton.clicked.connect(self.get_filepath)

        self.buttonBox.button(QtWidgets.QDialogButtonBox.Apply).clicked.connect(
            self.save
        )
        self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).clicked.connect(self.save)

        # TODO: disable OK and Apply if no path provided?

    def get_filepath(self):
        """Get export file path"""

        current_filepath = Path(self.filePathEdit.text()).resolve()
        current_filedir = current_filepath.parent

        # TODO: use QFileDialog.getOpenFileName?
        dialog = QtWidgets.QFileDialog(self)
        dialog.setDirectory(str(current_filedir))
        dialog.setFileMode(QtWidgets.QFileDialog.AnyFile)
        if dialog.exec_():
            self.filePathEdit.setText(dialog.selectedFiles()[0])

    def save(self):
        """Save ICS export settings"""

        # TODO: Validate settings before saving? (filepath writable?)

        self.settings.beginGroup("ics_export")
        self.settings.setValue("filepath", self.filePathEdit.text())
        self.settings.setValue("alarm", self.alarmsCheckBox.isChecked())
        self.settings.setValue("alarm_days", self.alarmsDelaySpinBox.value())
        self.settings.setValue(
            "custom_properties", self.customVeventTextEdit.toPlainText()
        )
        self.settings.setValue(
            "alarm_custom_properties", self.customValarmTextEdit.toPlainText()
        )
        self.settings.endGroup()


class PreferencesDialog(QtWidgets.QDialog):
    """Settings dialog"""

    def __init__(self, settings, main_window):

        super().__init__(main_window)

        uic.loadUi(str(UI_FILES_DIR / "preferencesdialog.ui"), self)

        self.settings = settings
        self.main_window = main_window

        self.pastSpinBox.setValue(self.settings.value("firstday", type=int))
        self.nextSpinBox.setValue(self.settings.value("lastday", type=int))
        self.icsExportCheckBox.setChecked(
            self.settings.value("ics_export/enabled", type=bool)
        )
        self.icsExportButton.setEnabled(
            self.settings.value("ics_export/enabled", type=bool)
        )

        self.icsExportCheckBox.stateChanged.connect(self.icsExportButton.setEnabled)
        self.icsExportButton.clicked.connect(
            lambda: IcsExportPreferencesDialog(self.settings, self).exec_()
        )

        self.bcknd_chkbx = {}

        for bcknd in BACKENDS:

            hbox = QtWidgets.QHBoxLayout()
            self.backendsLayout.addLayout(hbox)
            self.bcknd_chkbx[bcknd.id] = QtWidgets.QCheckBox(bcknd.name)
            hbox.addWidget(self.bcknd_chkbx[bcknd.id])

            if bcknd.cls is not None:
                bcknd_used = self.settings.value(bcknd.id + "/enabled", type=bool)
                self.bcknd_chkbx[bcknd.id].setChecked(bcknd_used)
                if bcknd.cls.CONFIG_DLG is not None:
                    button = QtWidgets.QPushButton(self.tr("Preferences"))
                    button.setEnabled(bcknd_used)
                    self.bcknd_chkbx[bcknd.id].stateChanged.connect(button.setEnabled)
                    button.clicked.connect(
                        # http://stackoverflow.com/questions/2295290/
                        # http://stackoverflow.com/questions/18836291/
                        lambda ignore, dlg=bcknd.cls.CONFIG_DLG: dlg(
                            self.settings, self
                        ).exec_()
                    )
                    hbox.addWidget(button)
            else:
                self.bcknd_chkbx[bcknd.id].setChecked(False)
                self.bcknd_chkbx[bcknd.id].setEnabled(False)
                self.bcknd_chkbx[bcknd.id].setToolTip(bcknd.exc_str)

        self.buttonBox.button(QtWidgets.QDialogButtonBox.Apply).clicked.connect(
            self.save
        )
        self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).clicked.connect(self.save)

    def save(self):
        """Save settings"""

        # TODO: Validate settings before saving?

        # Save settings
        self.settings.setValue("firstday", self.pastSpinBox.value())
        self.settings.setValue("lastday", self.nextSpinBox.value())
        self.settings.setValue("ics_export/enabled", self.icsExportCheckBox.isChecked())

        for bcknd in BACKENDS:
            self.settings.setValue(
                bcknd.id + "/enabled", self.bcknd_chkbx[bcknd.id].isChecked()
            )

        # Refresh birthday list
        self.main_window.reload()



"""Information about QBirthday"""

from PyQt5 import QtCore


__title__ = "QBirthday"
__summary__ = QtCore.QT_TR_NOOP("Birthday reminder")
__uri__ = "https://github.com/lafrech/qbirthday"

__version__ = "0.7.0b3"

__author__ = "Jérôme Lafréchoux"
__email__ = "jerome@jolimont.fr"

__authors__ = (
    "Andreas Bresser <andreas@phidev.org>",
    "Stefan Jurco <stefan.jurco@gmail.com>",
    "Alex Mallo <dernalis@gmail.com>",
    "Thomas Spura <tomspur@fedoraproject.org>",
    "Robert Wildburger <r.wildburger@gmx.at>",
    "Jérôme Lafréchoux <jerome@jolimont.fr>",
)
__artists__ = "Alex Mallo <dernalis@gmail.com>"

__copyright__ = (
    "2015 Jérôme Lafréchoux",
    "2009 Thomas Spura",
    "2009 Andreas Bresser",
    "2007 Alex Mallo",
)

__license__ = "GNU General Public License Version 2"
__license_long__ = """Licensed under the GNU General Public License Version 2

QBirthday is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

QBirthday is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301, USA."""



"""Backend exceptions"""

from qbirthday.exceptions import QBirthdayError


class BackendError(QBirthdayError):
    """Generic backend error"""


class BackendMissingLibraryError(BackendError):
    """Missing backend library"""


class BackendReadError(BackendError):
    """Error while reading from file/database"""


class BackendWriteError(BackendError):
    """Error while writing to file/database"""



"""Base class for all backends"""

import abc
import importlib

from PyQt5 import QtCore

from .exceptions import BackendMissingLibraryError


MISSING_LIB_ERR_STR = QtCore.QT_TRANSLATE_NOOP("Backends", "Missing {} library.")


class AbstractQObjectMetaclass(type(QtCore.QObject), abc.ABCMeta):
    """Metaclass for an abstract class inheriting QObject

    See https://stackoverflow.com/questions/46837947/
    """


class BaseBackend(QtCore.QObject, metaclass=AbstractQObjectMetaclass):
    """Abstract class for read-only backends

    To create a child class
    - create parse method
    - override class attributes if needed
    """

    # Configuration dialog
    CONFIG_DLG = None
    # Default configuration values
    DEFAULTS = {}

    def __init__(self, settings):
        super().__init__()
        self.settings = settings

    @abc.abstractmethod
    def parse(self):
        """Return birthdates as list of (name, date) tuples"""


def try_import(module_name, lib_name=None):
    """Import module

    In case of ImportError, raise BackendMissingLibraryError
    module_name: name of the module to import
    lib_name: name of the python lib containing the module
        (default to module_name)
    """
    lib_name = lib_name or module_name
    try:
        module = importlib.import_module(module_name)
    except ImportError:
        import_err = QtCore.QCoreApplication.translate(
            "Backends", MISSING_LIB_ERR_STR
        ).format(lib_name)
        raise BackendMissingLibraryError(import_err)
    return module



"""Main window"""

import datetime

from PyQt5 import QtCore, QtGui, QtWidgets, uic

from .birthday_list import BirthdayList
from .statusicon import StatusIcon
from .settings import Settings
from .paths import PICS_PATHS, UI_FILES_DIR


class MainWindow(QtWidgets.QMainWindow):
    """Main window"""

    def __init__(self):

        super().__init__()

        self.setWindowFlags(
            QtCore.Qt.Tool
            | QtCore.Qt.FramelessWindowHint
            | QtCore.Qt.WindowStaysOnTopHint
        )  # ??

        uic.loadUi(str(UI_FILES_DIR / "mainwindow.ui"), self)

        self.settings = Settings()
        self.bday_list = BirthdayList(self, self.settings)
        self.status_icon = StatusIcon(self, self.settings)

        # Initialise current day
        self.current_day = datetime.datetime.now().strftime("%d")

        # Check every 60 seconds for new day
        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.check_new_day)
        self.timer.start(60 * 1000)

        self.reload()

        # TODO: Catch mouse focus out and close window

    def showEvent(self, event):
        """Callback for "show" event"""

        super().showEvent(event)

        # Window shall appear under system tray icon
        systray_icon_pos = self.status_icon.geometry().center()
        self.move(systray_icon_pos.x() - self.width() // 2, systray_icon_pos.y())

        # Ensure the window is not minimized on virtual desktop change
        self.showNormal()

    def reload(self):
        """Reload birthdays and update GUI accordingly"""

        # Reload birthdays
        self.bday_list.reload()

        # Reload status icon
        self.status_icon.reload_set_icon()

        # Set title
        if self.bday_list.bdays_in_period():
            self.titleLabel.setText(self.tr("Birthdays"))
        else:
            self.titleLabel.setText(self.tr("No birthdays in specified period"))

        # Empty birthday list
        # http://stackoverflow.com/questions/4528347/
        for i in reversed(range(self.birthdaysLayout.count())):
            widget = self.birthdaysLayout.itemAt(i).widget()
            self.birthdaysLayout.removeWidget(widget)
            widget.setParent(None)

        # Add birthdays
        firstday = self.settings.value("firstday", type=int)
        lastday = self.settings.value("lastday", type=int)
        for delta_day in range(firstday, lastday + 1):
            for birthdate, name in self.bday_list.check_day(delta_day):
                row = self.birthdaysLayout.rowCount()
                for col, label in enumerate(
                    self.make_birthday_line(delta_day, name, birthdate)
                ):
                    self.birthdaysLayout.addWidget(label, row, col)

    def check_new_day(self):
        """Check for new day

        Should be called e.g. every 60 seconds to check if day has changed.
        """
        new_day = datetime.datetime.now().strftime("%d")
        if self.current_day != new_day:
            self.current_day = new_day
            self.reload()

    def make_birthday_line(self, delta_day, name, birthdate):
        """Return a row as a list of widgets

        deltaday (int): day index in specified period
        name (str): person name
        birthdate (dt.date): date of birth
        """
        label_image = QtWidgets.QLabel()
        # TODO: use Qt to localize day + month
        label_day = QtWidgets.QLabel(str(birthdate.day))
        label_month = QtWidgets.QLabel(birthdate.strftime("%B"))
        label_name = QtWidgets.QLabel(name)
        label_when = QtWidgets.QLabel()
        label_age = QtWidgets.QLabel(
            self.tr("{} years old").format(datetime.date.today().year - birthdate.year)
        )

        label_day.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        label_when.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        label_age.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)

        labels = [
            label_image,
            label_day,
            label_month,
            label_name,
            label_when,
            label_age,
        ]

        # Birthday today
        if delta_day == 0:
            label_image.setPixmap(QtGui.QPixmap(PICS_PATHS["birthdaytoday"]))
            label_when.setText(self.tr("Today"))
            for label in labels:
                label.setStyleSheet("QLabel { font: bold; }")
        # Birthday in the past
        elif delta_day < 0:
            label_image.setPixmap(QtGui.QPixmap(PICS_PATHS["birthdaylost"]))
            if delta_day == -1:
                label_when.setText(self.tr("Yesterday"))
            else:
                label_when.setText(self.tr("{} days ago").format(-delta_day))
            for label in labels:
                label.setStyleSheet("QLabel { color : grey; }")
        # Birthday in the future
        else:
            label_image.setPixmap(QtGui.QPixmap(PICS_PATHS["birthdaynext"]))
            if delta_day == 1:
                label_when.setText(self.tr("Tomorrow"))
            else:
                label_when.setText(self.tr("{} days").format(delta_day))

        return labels

    def show_error_message(self, msg):
        """Display an error message in a dialog box"""
        QtWidgets.QMessageBox.warning(
            self,
            QtCore.QCoreApplication.applicationName(),
            msg,
            QtWidgets.QMessageBox.Close,
        )



"""About dialog"""

from PyQt5 import QtGui, QtWidgets, uic

from qbirthday import __about__ as about
from .paths import PICS_PATHS, UI_FILES_DIR


class AboutDialog(QtWidgets.QDialog):
    """About Dialog"""

    def __init__(self, main_window):

        super().__init__(main_window)

        uic.loadUi(str(UI_FILES_DIR / "aboutdialog.ui"), self)

        # About
        self.imageLabel.setPixmap(QtGui.QPixmap(PICS_PATHS["qbirthday"]))
        self.nameLabel.setText(f"{about.__title__} {about.__version__}")
        self.descLabel.setText(self.tr(about.__summary__))
        self.copyrightLabel.setText(
            "Copyright ©\n{}".format("\n".join(about.__copyright__))
        )
        self.uriLabel.setText("<a href={uri}>{uri}</a>".format(uri=about.__uri__))

        # Authors
        self.authorsTextEdit.insertPlainText("\n".join(about.__authors__))

        # Translators
        translators = self.tr("translator-credit")
        if translators == "translator-credit":
            translators = self.tr(
                "There are no translations or the translator "
                "doesn't want to get credits for that."
            )
        self.translatorsTextEdit.insertPlainText(translators)

        # Artists
        self.artistsTextEdit.insertPlainText(about.__artists__)

        # License
        self.licenseTextEdit.insertPlainText(about.__license_long__)
        self.licenseTextEdit.moveCursor(QtGui.QTextCursor.Start)



"""Lightning backend"""

import configparser
import datetime as dt
from pathlib import Path
import sqlite3

from .base import BaseBackend
from .exceptions import BackendReadError


BACKEND_ID = "Lightning"
BACKEND_NAME = "Lightning"


class LightningBackend(BaseBackend):
    """Lightning backend

    The user should add birthdates with
    - Category: Birthday
    - Start date: Birth date (year matters for age calculation)
    - Title: Name of the person
    The Repeat parameter does not matter
    """

    @staticmethod
    def _parse_profile(filepath):
        try:
            with sqlite3.connect(str(filepath)) as conn:
                cursor = conn.cursor()
                qry = """SELECT title, event_start FROM cal_events ce
                      INNER JOIN cal_properties cp
                      ON ce.id = cp.item_id
                      WHERE cp.key == 'CATEGORIES' AND
                      cp.value == 'Birthday' AND
                      ce.title != '';"""
                cursor.execute(qry)
                birthdates = [
                    (name, dt.datetime.utcfromtimestamp(int(bdate_ts) / 1000000).date())
                    for name, bdate_ts in cursor
                ]
                return birthdates
        except sqlite3.Error as exc:
            raise BackendReadError(exc)

    def parse(self):
        """Parse Lightning sqlite database in Thunderbird profile directory"""
        profiles_dir = Path.home() / ".thunderbird"
        if not profiles_dir.is_dir():
            profiles_dir = Path.home() / ".mozilla-thunderbird"
            if not profiles_dir.is_dir():
                raise BackendReadError(
                    self.tr("Cannot find Lightning profile directory")
                )
        profilefile = profiles_dir / "profiles.ini"
        if profilefile.is_file():
            conf_pars = configparser.ConfigParser()
            conf_pars.read(str(profilefile))
            profiles = {}  # dict of founded profiles
            # get profiles from profiles.ini in thunderbird directory
            for sec in conf_pars.sections():
                if sec.lower().startswith("profile"):
                    profiles[sec] = {}
                    for opt in conf_pars.options(sec):
                        profiles[sec][opt.lower()] = conf_pars.get(sec, opt)
            # get all data from profiles.ini
            for profile in profiles:
                if profiles[profile]["isrelative"]:
                    profile_location = profiles_dir / profiles[profile]["path"]
                else:
                    profile_location = profiles[profile]["path"]
                db_location = profile_location / "calendar-data/local.sqlite"
                return self._parse_profile(db_location)
        # Missing profile file
        raise BackendReadError(
            self.tr("Error reading profile file {}").format(str(profilefile))
        )


BACKEND = LightningBackend



"""Module defining paths

PICS_DIR, UI_FILES_DIR and QM_FILES_DIR are Path instances
They must be cast as str when fed to Qt objects

PICS_PATHS is a dict of strings
"""

from pathlib import Path

from PyQt5 import QtCore


GENERIC_DATA_LOCATION = (
    Path(
        QtCore.QStandardPaths.writableLocation(
            QtCore.QStandardPaths.GenericDataLocation
        )
    )
    / "QBirthday"
)
APP_DATA_LOCATION = (
    Path(QtCore.QStandardPaths.writableLocation(QtCore.QStandardPaths.AppDataLocation))
    / "QBirthday"
)

# Create default data storage directories
GENERIC_DATA_LOCATION.mkdir(exist_ok=True)
APP_DATA_LOCATION.mkdir(exist_ok=True)


PICS_DIR = Path(__file__).parent / "pics"

PICS_PATHS = {
    "birthdaylost": str(PICS_DIR / "birthdaylost.png"),
    "birthdaynext": str(PICS_DIR / "birthdaynext.png"),
    "birthday": str(PICS_DIR / "birthday.png"),
    "birthdayred": str(PICS_DIR / "birthdayred.png"),
    "birthdaytoday": str(PICS_DIR / "birthdaytoday.png"),
    "qbirthday": str(PICS_DIR / "qbirthday.png"),
    "nobirthday": str(PICS_DIR / "nobirthday.png"),
}

UI_FILES_DIR = Path(__file__).parent / "ui"

QM_FILES_DIR = Path(__file__).parent / "i18n"



"""Birthday backends"""

import importlib
from collections import namedtuple

from .exceptions import BackendMissingLibraryError


# TODO: translate display names

# Backend description tuple
# id: short name used in settings file
# name: translatable display name
# cls: backend class or None is ImportError
# exc_str: translatable import error message or None
BackendDescription = namedtuple("BackendDescription", ["id", "name", "cls", "exc_str"])

BACKENDS = []

for module_name in ("csv", "mysql", "lightning"):
    try:
        module = importlib.import_module(f"qbirthday.backends.{module_name}")
    except BackendMissingLibraryError as exc:
        BACKENDS.append(
            BackendDescription(exc.bcknd_id, exc.bcknd_name, None, str(exc))
        )
    else:
        bcknd_id = getattr(module, "BACKEND_ID")
        bcknd_name = getattr(module, "BACKEND_NAME")
        bcknd_cls = getattr(module, "BACKEND")
        BACKENDS.append(BackendDescription(bcknd_id, bcknd_name, bcknd_cls, None))



"""MySQL backend"""

from PyQt5 import QtWidgets, uic

from qbirthday.paths import UI_FILES_DIR
from .base import BaseBackend, try_import
from .exceptions import BackendMissingLibraryError, BackendReadError


BACKEND_ID = "MySQL"
BACKEND_NAME = "MySQL database"


try:
    # pylint: disable=invalid-name
    MySQLdb = try_import("MySQLdb", "mysqlclient")
except BackendMissingLibraryError as exc:
    exc.bcknd_id = BACKEND_ID
    exc.bcknd_name = BACKEND_NAME
    raise exc


class MySqlPreferencesDialog(QtWidgets.QDialog):
    """MySQL backend settings dialog"""

    def __init__(self, settings, parent):

        super().__init__(parent)

        uic.loadUi(str(UI_FILES_DIR / "mysqlpreferencesdialog.ui"), self)

        self.settings = settings

        # Fill fields with current values
        self.settings.beginGroup("MySQL")
        self.hostEdit.setText(self.settings.value("host", type=str))
        self.portEdit.setText(str(self.settings.value("port", type=int)))
        self.usernameEdit.setText(self.settings.value("username", type=str))
        self.passwordEdit.setText(self.settings.value("password", type=str))
        self.databaseEdit.setText(self.settings.value("database", type=str))
        self.tableEdit.setText(self.settings.value("table", type=str))
        self.nameColEdit.setText(self.settings.value("namecol", type=str))
        self.dateColEdit.setText(self.settings.value("datecol", type=str))
        self.settings.endGroup()

        self.buttonBox.button(QtWidgets.QDialogButtonBox.Apply).clicked.connect(
            self.save
        )
        self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).clicked.connect(self.save)

        # TODO: disable OK and Apply if empty/invalid field

    def save(self):
        """Save MySQL backend settings"""

        self.settings.beginGroup("MySQL")
        self.settings.setValue("host", self.hostEdit.text())
        self.settings.setValue("port", int(self.portEdit.text()))
        self.settings.setValue("username", self.usernameEdit.text())
        self.settings.setValue("password", self.passwordEdit.text())
        self.settings.setValue("database", self.databaseEdit.text())
        self.settings.setValue("table", self.tableEdit.text())
        self.settings.setValue("namecol", self.nameColEdit.text())
        self.settings.setValue("datecol", self.dateColEdit.text())
        self.settings.endGroup()


class MySQLBackend(BaseBackend):
    """MySQL backend"""

    CONFIG_DLG = MySqlPreferencesDialog

    DEFAULTS = {
        "host": "localhost",
        "port": 3306,
        "username": "",
        "password": "",
        "database": "",
        "table": "person",
        "namecol": "name",
        "datecol": "date",
    }

    def __init__(self, settings):

        super().__init__(settings)

        self.settings.beginGroup("MySQL")
        self.host = self.settings.value("host", type=str)
        self.port = self.settings.value("port", type=int)
        self.username = self.settings.value("username", type=str)
        self.password = self.settings.value("password", type=str)
        self.database = self.settings.value("database", type=str)
        self.table = self.settings.value("table", type=str)
        self.name_col = self.settings.value("namecol", type=str)
        self.date_col = self.settings.value("datecol", type=str)
        self.settings.endGroup()

        self.cursor = None
        self.conn = None

    def _connect(self):
        """establish connection"""

        try:
            self.conn = MySQLdb.connect(
                host=self.host,
                port=self.port,
                user=self.username,
                passwd=self.password,
                db=self.database,
            )
            self.cursor = self.conn.cursor()
        except Exception as exc:
            raise ConnectionError(exc)

    def parse(self):
        """connect to mysql-database and get data"""

        # TODO: use a context manager
        try:
            self._connect()
        except ConnectionError as exc:
            raise BackendReadError(exc)

        try:
            qry = f"SELECT {self.name_col}, {self.date_col} FROM {self.table}"
            self.cursor.execute(qry)
            rows = self.cursor.fetchall()
            birthdates = []
            for row in rows:
                birthdates.append((row[0], row[1]))
        except Exception as msg:
            raise BackendReadError(
                self.tr("Could not execute SQL query '{}':\n{}").format(qry, msg)
            )
        else:
            return birthdates
        finally:
            self.conn.close()


BACKEND = MySQLBackend
